<?php

namespace Drupal\easy_entity_field\Form;

use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Field\FallbackFieldTypeCategory;
use Drupal\field\Entity\FieldStorageConfig;
use Drupal\field_ui\Form\FieldStorageAddForm;
use Drupal\Core\Form\FormStateInterface;

/**
 * Provides a form for the "easy entity field" add page.
 *
 * @internal
 */
class EasyEntityFieldAddForm extends FieldStorageAddForm {

  /**
   * The name of the entity type.
   *
   * @var string
   */
  protected $entityTypeId;

  /**
   * Save field type definitions and categories in the form state.
   *
   * Get all field type definitions and store each one twice:
   * - field_type_options: each field type is indexed by its category plugin ID
   *   or its label.
   * - unique_definitions: each field type is indexed by its category and name.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  protected function processFieldDefinitions(FormStateInterface $form_state): void {
    // Fields to be removed in the option list.
    $field_type_removed = ['path' => TRUE];

    // Fields to be removed from the hook.
    $hook = 'easy_entity_field_remove_field_type';
    if (\Drupal::moduleHandler()->hasImplementations($hook)) {
      $field_type_removed = NestedArray::mergeDeep($field_type_removed, \Drupal::moduleHandler()->invokeAll($hook, [$field_type_removed]));
    }

    $field_type_options = $unique_definitions = [];
    $definitions = $this->fieldTypePluginManager->getUiDefinitions() + $this->fieldTypePluginManager->getDefinitions();

    $grouped_definitions = $this->fieldTypePluginManager->getGroupedDefinitions($definitions, 'label', 'id');

    foreach ($grouped_definitions as $category => $field_types) {
      foreach ($field_types as $name => $field_type) {
        // Remove fields from the option list.
        if (!empty($field_type_removed[$name])) {
          continue;
        }

        $definition = ['unique_identifier' => $name] + $field_type;
        $category_info = $this->fieldTypeCategoryManager->createInstance($field_type['category'], $definition);
        $definition['display_as_group'] = !($category_info instanceof FallbackFieldTypeCategory);
        $definition['no_ui'] = $field_type['no_ui'];
        $id = $this->fieldTypeCategoryManager->hasDefinition($category)
          ? $category_info->getPluginId()
          : (string) $field_type['label'];
        $field_type_options[$id] = $definition;
        $unique_definitions[$category][$name] = $definition;
      }
    }

    $form_state->set('field_type_options', $field_type_options);
    $form_state->set('unique_definitions', $unique_definitions);
  }

  /**
   * Adds field types for the selected group to the form.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  protected function addFieldOptionsForGroup(array &$form, FormStateInterface $form_state): void {
    // Field label and field_name.
    parent::addFieldOptionsForGroup($form, $form_state);

    $form['new_storage_wrapper']['field_name'] = [
      '#type' => 'machine_name',
      '#size' => 15,
      '#description' => $this->t('A unique machine-readable name containing letters, numbers, and underscores.'),
      // Calculate characters depending on the length of the field prefix
      // setting. Maximum length is 32.
      '#maxlength' => FieldStorageConfig::NAME_MAX_LENGTH,
      '#machine_name' => [
        'source' => ['new_storage_wrapper', 'label'],
        'exists' => [$this, 'fieldNameExists'],
      ],
      '#required' => FALSE,
    ];

    $field_type_options = $form_state->get('field_type_options');
    $new_storage_type = $form_state->getValue('new_storage_type');
    $form['no_ui'] = [
      '#type' => 'value',
      '#value' => $field_type_options[$new_storage_type]['no_ui'],
    ];

    if (empty($field_type_options[$new_storage_type]['no_ui'])) {
      $form['form_display'] = [
        '#type' => 'checkbox',
        '#title' => $this->t('Create form display'),
        '#default_value' => TRUE,
      ];

      $form['view_display'] = [
        '#type' => 'checkbox',
        '#title' => $this->t('Create view display'),
        '#default_value' => TRUE,
      ];
    }
  }

  /**
   * Validates the second step (field storage selection and label) of the form.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  public function validateFieldType(array $form, FormStateInterface $form_state) {
    // Missing label.
    if (!$form_state->getValue('label')) {
      $form_state->setErrorByName('label', $this->t('Add new field: you need to provide a label.'));
    }
    // Missing field name.
    if (!$form_state->getValue('field_name')) {
      $form_state->setErrorByName('field_name', $this->t('Add new field: you need to provide a machine name for the field.'));
    }
    // Field name validation.
    else {
      $field_name = $form_state->getValue('field_name');
      $form_state->setValueForElement($form['new_storage_wrapper']['field_name'], $field_name);
    }

    if (isset($form['group_field_options_wrapper']['fields']) && !$form_state->getValue('group_field_options_wrapper')) {
      $form_state->setErrorByName('group_field_options_wrapper', $this->t('You need to choose an option.'));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state): void {
    $values = $form_state->getValues();

    $entity_type = $this->entityTypeManager->getDefinition($this->entityTypeId);
    $field_storage_type = $values['group_field_options_wrapper'] ?? $values['new_storage_type'];

    // Create new field.
    if ($field_storage_type) {
      $field_values = [
        'label' => $values['label'],
        'field_name' => $values['field_name'],
        'entity_type' => $this->entityTypeId,
        'field_type' => $field_storage_type,
        'translatable' => $entity_type->isTranslatable(),
      ];

      $widget_id = $formatter_id = '';
      $widget_settings = $formatter_settings = [];
      $storage_settings = $field_settings = [];
      // Check if we're dealing with a preconfigured field.
      if (str_starts_with($field_storage_type, 'field_ui:')) {
        [, $field_type, $option_key] = explode(':', $field_values['field_type'], 3);
        $field_values['field_type'] = $field_type;

        $field_definition = $this->fieldTypePluginManager->getDefinition($field_type);
        $options = $this->fieldTypePluginManager->getPreconfiguredOptions($field_definition['id']);
        $field_options = $options[$option_key];

        // Merge in preconfigured field storage options.
        if (isset($field_options['field_storage_config'])) {
          $storage_settings = $field_options['field_storage_config']['settings'];
        }

        // Merge in preconfigured field options.
        if (isset($field_options['field_config'])) {
          $field_values['required'] = $field_options['field_config']['required'];
          $field_settings = $field_options['field_config']['settings'];
        }

        $widget_id = $field_options['entity_form_display']['type'] ?? '';
        $widget_settings = $field_options['entity_form_display']['settings'] ?? [];
        $formatter_id = $field_options['entity_view_display']['type'] ?? '';
        $formatter_settings = $field_options['entity_view_display']['settings'] ?? [];
      }
      else {
        if (!empty($this->fieldTypePluginManager->getDefaultStorageSettings($field_storage_type))) {
          $storage_settings = $this->fieldTypePluginManager->getDefaultStorageSettings($field_storage_type);
        }

        if (!empty($this->fieldTypePluginManager->getDefaultFieldSettings($field_storage_type))) {
          $field_settings = $this->fieldTypePluginManager->getDefaultFieldSettings($field_storage_type);
        }
      }

      $field_values['display']['form'] = !empty($values['form_display']);
      $field_values['display']['formatter_type'] = $formatter_id;
      $field_values['display']['formatter_settings'] = $formatter_settings;

      $field_values['display']['view'] = !empty($values['view_display']);
      $field_values['display']['widget_type'] = $widget_id;
      $field_values['display']['widget_settings'] = $widget_settings;

      $field_values['field_settings'] = $field_settings;
      $field_values['storage_settings'] = $storage_settings;

      // Create the field storage and field.
      try {
        $field = $this->entityTypeManager->getStorage('easy_entity_field')->create($field_values);
        $field->save();

        if (!empty($values['no_ui'])) {
          $form_state->setRedirectUrl($field->toUrl('edit-form'));
        }
        else {
          $form_state->setRedirectUrl($field->toUrl('storage-form'));
        }
      }
      catch (\Exception $e) {
        $this->messenger()->addError($this->t('There was a problem creating field %label: @message', [
          '%label' => $values['label'],
          '@message' => $e->getMessage(),
        ]));
      }
    }

    // Store new field information for any additional submit handlers.
    $form_state->set(['fields_added', '_add_new_field'], $values['field_name']);
  }

  /**
   * Checks if a field machine name is taken.
   *
   * @param string $value
   *   The machine name, not prefixed.
   * @param array $element
   *   An array containing the structure of the 'field_name' element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @return bool
   *   Whether or not the field machine name is taken.
   */
  public function fieldNameExists($value, $element, FormStateInterface $form_state): bool {
    $field_storage_definitions = $this->entityFieldManager->getFieldStorageDefinitions($this->entityTypeId);
    return isset($field_storage_definitions[$value]);
  }

}
